#include "batch.h"
#include "buffer.h"
#include "camera.h"
#include "clipping.h"
#include "error.h"
#include "lighting.h"
#include "material.h"
#include "program.h"
#include "programdata.h"
#include "renderable.h"
#include "renderer.h"
#include "sampler.h"
#include "texture.h"
#include "texturing.h"
#include "texunit.h"
#include "vertexarray.h"
#include "vertexsetup.h"
#include "windingtest.h"

using namespace std;

namespace Msp {
namespace GL {

Renderer::Renderer():
	default_camera(0)
{
	init();
}

Renderer::Renderer(const Camera *c):
	default_camera(c)
{
	init();

	if(c)
		set_camera(*c);
}

void Renderer::init()
{
	state_stack.reserve(16);
	state_stack.push_back(State());
	shdata_stack.reserve(32);
	state = &state_stack.back();

	standard_shdata.uniform("projection_matrix", Matrix());
	standard_shdata.uniform("eye_world_matrix", Matrix());
}

Renderer::~Renderer()
{
	end();
}

void Renderer::set_camera(const Camera &c)
{
	state->camera = &c;
	standard_shdata.uniform("projection_matrix", state->camera->get_projection_matrix());
	standard_shdata.uniform("eye_world_matrix", state->camera->get_view_matrix());
	changed |= STANDARD_SHDATA;
	set_matrix(state->camera->get_view_matrix());
}

void Renderer::set_matrix(const Matrix &matrix)
{
	state->modelview_matrix = matrix;
	changed |= MATRIX;
}

void Renderer::transform(const Matrix &matrix)
{
	state->modelview_matrix *= matrix;
	changed |= MATRIX;
}

void Renderer::set_texture(Tag tag, const Texture *tex, const Sampler *samp)
{
	set_texture(tag, -1, tex, samp);
}

void Renderer::set_texture(Tag tag, int unit, const Texture *tex, const Sampler *samp)
{
	if(texture_stack.size()>state->texture_count)
	{
		BoundTexture &bt = texture_stack[state->texture_count];
		if((!tag.id || bt.tag==tag) && (unit<0 || bt.unit==unit) && bt.texture==tex && bt.sampler==samp)
		{
			++state->texture_count;
			return;
		}
		else
			flush_textures();
	}

	for(vector<BoundTexture>::iterator i=texture_stack.end(); i!=texture_stack.begin(); )
		if((--i)->tag==tag && i->unit==unit)
		{
			i->replaced = texture_stack.size();
			break;
		}

	texture_stack.push_back(BoundTexture());
	BoundTexture &bound_tex = texture_stack.back();
	bound_tex.tag = tag;
	bound_tex.unit = unit;
	bound_tex.texture = tex;
	bound_tex.sampler = samp;
	state->texture_count = texture_stack.size();
}

void Renderer::set_texture(const Texture *t, const Sampler *s)
{
	set_texture(Tag(), 0, t, s);
}

void Renderer::flush_textures()
{
	for(unsigned i=0; i<texture_stack.size(); ++i)
	{
		BoundTexture &bt = texture_stack[i];
		if(i>=state->texture_count && bt.unit>=0)
		{
			Texture::unbind_from(bt.unit);
			Sampler::unbind_from(bt.unit);
		}
		else if(bt.replaced>=static_cast<int>(state->texture_count))
			bt.replaced = -1;
	}

	texture_stack.erase(texture_stack.begin()+state->texture_count, texture_stack.end());
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
void Renderer::set_texturing(const Texturing *t)
{
	if(t)
	{
		unsigned n_units = TexUnit::get_n_units();
		for(unsigned i=0; i<n_units; ++i)
			if(const Texture *tex = t->get_attached_texture(i))
				set_texture(Tag(), i, tex, t->get_attached_sampler(i));
	}
}
#pragma GCC diagnostic pop

unsigned Renderer::allocate_effect_texunit()
{
	return --state->lowest_effect_texunit;
}

void Renderer::set_material(const Material *m)
{
	state->material = m;
	changed |= MATERIAL_SHDATA;
}

void Renderer::set_lighting(const Lighting *l)
{
	state->lighting = l;
	state->lighting_matrix = state->modelview_matrix;
	if(l)
	{
		l->update_shader_data(standard_shdata, state->lighting_matrix);
		changed |= STANDARD_SHDATA;
	}
}

void Renderer::set_clipping(const Clipping *c)
{
	state->clipping = c;
	state->clipping_matrix = state->modelview_matrix;
	if(c)
	{
		c->update_shader_data(standard_shdata, state->clipping_matrix);
		changed |= STANDARD_SHDATA;
	}
}

void Renderer::set_shader_program(const Program *p, const ProgramData *d)
{
	state->shprog = p;
	if(p && d)
		add_shader_data(*d);
}

void Renderer::add_shader_data(const ProgramData &d)
{
	if(state->shdata_count<shdata_stack.size() && shdata_stack[state->shdata_count]==&d)
		++state->shdata_count;
	else
	{
		flush_shader_data();
		shdata_stack.push_back(&d);
		state->shdata_count = shdata_stack.size();
		changed |= SHADER_DATA;
	}
}

void Renderer::flush_shader_data()
{
	if(shdata_stack.size()>state->shdata_count)
		shdata_stack.erase(shdata_stack.begin()+state->shdata_count, shdata_stack.end());
}

void Renderer::set_vertex_setup(const VertexSetup *vs)
{
	state->vertex_setup = vs;
}

void Renderer::set_winding_test(const WindingTest *w)
{
	state->winding_test = w;
}

void Renderer::set_reverse_winding(bool r)
{
	state->reverse_winding = r;
}

void Renderer::set_object_lod_bias(unsigned b)
{
	state->object_lod_bias = b;
}

void Renderer::push_state()
{
	state_stack.push_back(state_stack.back());
	state = &state_stack.back();
}

void Renderer::pop_state()
{
	if(state_stack.size()==1)
		throw stack_underflow("Renderer::pop_state");

	const Camera *old_camera = state->camera;
	const Lighting *old_lighting = state->lighting;
	const Clipping *old_clipping = state->clipping;
	state_stack.pop_back();
	state = &state_stack.back();
	changed |= MATRIX;
	bool camera_changed = (state->camera!=old_camera);
	if(camera_changed)
	{
		if(state->camera)
		{
			standard_shdata.uniform("projection_matrix", state->camera->get_projection_matrix());
			standard_shdata.uniform("eye_world_matrix", state->camera->get_view_matrix());
		}
		else
		{
			standard_shdata.uniform("projection_matrix", Matrix());
			standard_shdata.uniform("eye_world_matrix", Matrix());
		}
		changed |= STANDARD_SHDATA;
	}
	/* This actually should compare the relevant matrices rather than check for
	camera, but in practice lighting and clipping is set right after the camera
	and a boolean check is much faster than matrix comparison. */
	if(state->lighting!=old_lighting || camera_changed)
	{
		if(state->lighting)
		{
			state->lighting->update_shader_data(standard_shdata, state->lighting_matrix);
			changed |= STANDARD_SHDATA;
		}
	}
	if(state->clipping!=old_clipping || camera_changed)
	{
		if(state->clipping)
		{
			state->clipping->update_shader_data(standard_shdata, state->clipping_matrix);
			changed |= STANDARD_SHDATA;
		}
	}
}

void Renderer::end()
{
	if(state_stack.size()>1)
		throw invalid_operation("Renderer::end");

	*state = State();
	if(default_camera)
		set_camera(*default_camera);
	else
		standard_shdata.uniform("projection_matrix", Matrix());
	shdata_stack.clear();
	excluded.clear();

	for(vector<BoundTexture>::iterator i=texture_stack.begin(); i!=texture_stack.end(); ++i)
		if(i->unit>=0)
		{
			Texture::unbind_from(i->unit);
			Sampler::unbind_from(i->unit);
		}
	Clipping::unbind();
	Program::unbind();
	VertexSetup::unbind();
	Buffer::unbind_from(ELEMENT_ARRAY_BUFFER);
	WindingTest::unbind();
}

void Renderer::exclude(const Renderable &renderable)
{
	excluded.insert(&renderable);
}

void Renderer::include(const Renderable &renderable)
{
	excluded.erase(&renderable);
}

void Renderer::render(const Renderable &renderable, Tag tag)
{
	if(!excluded.count(&renderable))
		renderable.render(*this, tag);
}

void Renderer::draw(const Batch &batch)
{
	apply_state();

	batch.draw();
}

void Renderer::draw_instanced(const Batch &batch, unsigned count)
{
	apply_state();

	batch.draw_instanced(count);
}

void Renderer::apply_state()
{
	if(!state->shprog)
		throw invalid_operation("Renderer::apply_state");

	/* We (mostly) let the objects themselves figure out if the binding has
	changed */

	if(state->texture_count<texture_stack.size())
		flush_textures();

	for(vector<BoundTexture>::const_iterator i=texture_stack.begin(); i!=texture_stack.end(); ++i)
	{
		int unit = (i->tag.id ? state->shprog->get_uniform_binding(i->tag) : i->unit);
		if(unit>=0)
		{
			if(i->texture)
				i->texture->bind_to(unit);
			if(i->sampler)
				i->sampler->bind_to(unit);
			i->unit = unit;
		}
	}

	if(state->clipping)
		state->clipping->bind();
	else
		Clipping::unbind();

	bool shprog_changed = (state->shprog!=Program::current());
	state->shprog->bind();

	if(changed&MATRIX)
	{
		standard_shdata.uniform("eye_obj_matrix", state->modelview_matrix);
		LinAl::SquareMatrix<float, 3> nm = state->modelview_matrix.block<3, 3>(0, 0);
		nm = transpose(invert(nm));
		standard_shdata.uniform_matrix3("eye_obj_normal_matrix", &nm(0, 0));
		changed = (changed&~MATRIX)|STANDARD_SHDATA;
	}

	if(state->material && ((changed&MATERIAL_SHDATA) || shprog_changed))
	{
		state->material->get_shader_data().apply();
		changed &= ~MATERIAL_SHDATA;
	}

	if((changed&STANDARD_SHDATA) || shprog_changed)
	{
		standard_shdata.apply();
		changed &= ~STANDARD_SHDATA;
	}

	bool extra_shdata = (shdata_stack.size()>state->shdata_count);

	if((changed&SHADER_DATA) || shprog_changed || extra_shdata)
	{
		if(extra_shdata)
			shdata_stack.erase(shdata_stack.begin()+state->shdata_count, shdata_stack.end());
		for(vector<const ProgramData *>::const_iterator i=shdata_stack.begin(); i!=shdata_stack.end(); ++i)
			(*i)->apply();
		changed &= ~SHADER_DATA;
	}

	if(state->vertex_setup)
		state->vertex_setup->bind();
	else
		VertexSetup::unbind();

	if(state->winding_test)
	{
		if(state->reverse_winding)
			state->winding_test->get_reverse().bind();
		else
			state->winding_test->bind();
	}
	else
		WindingTest::unbind();
}


Renderer::BoundTexture::BoundTexture():
	unit(-1),
	texture(0),
	sampler(0),
	replaced(-1)
{ }


Renderer::State::State():
	camera(0),
	texture_count(0),
	lowest_effect_texunit(TexUnit::get_n_units()),
	material(0),
	lighting(0),
	clipping(0),
	shprog(0),
	shdata_count(0),
	vertex_setup(0),
	winding_test(0),
	reverse_winding(false),
	object_lod_bias(0)
{ }

} // namespace GL
} // namespace Msp
